@@ -12,10 +12,7 @@ pub mod rejection;
 pub mod ws;
 
 mod content_length_limit;
-mod extension;
-mod form;
-mod matched_path;
-mod query;
+mod host;
 mod raw_query;
 mod request_parts;
 
@@ -23,24 +20,37 @@ mod request_parts;
 pub use axum_core::extract::{FromRequest, RequestParts};
 
 #[doc(inline)]
-#[allow(deprecated)]
 pub use self::{
     connect_info::ConnectInfo,
     content_length_limit::ContentLengthLimit,
-    extension::Extension,
     extractor_middleware::extractor_middleware,
-    form::Form,
-    matched_path::MatchedPath,
+    host::Host,
     path::Path,
-    query::Query,
     raw_query::RawQuery,
-    request_parts::OriginalUri,
     request_parts::{BodyStream, RawBody},
 };
+
 #[doc(no_inline)]
 #[cfg(feature = "json")]
 pub use crate::Json;
 
+#[doc(no_inline)]
+pub use crate::Extension;
+
+#[cfg(feature = "form")]
+mod form;
+
+#[cfg(feature = "form")]
+#[doc(inline)]
+pub use self::form::Form;
+
+#[cfg(feature = "matched-path")]
+mod matched_path;
+
+#[cfg(feature = "matched-path")]
+#[doc(inline)]
+pub use self::matched_path::MatchedPath;
+
 #[cfg(feature = "multipart")]
 pub mod multipart;
 
@@ -48,38 +58,42 @@ pub mod multipart;
 #[doc(inline)]
 pub use self::multipart::Multipart;
 
+#[cfg(feature = "query")]
+mod query;
+
+#[cfg(feature = "query")]
+#[doc(inline)]
+pub use self::query::Query;
+
+#[cfg(feature = "original-uri")]
+#[doc(inline)]
+pub use self::request_parts::OriginalUri;
+
 #[cfg(feature = "ws")]
 #[doc(inline)]
 pub use self::ws::WebSocketUpgrade;
 
 #[cfg(feature = "headers")]
-mod typed_header;
-
-#[cfg(feature = "headers")]
-#[doc(inline)]
-pub use self::typed_header::TypedHeader;
+#[doc(no_inline)]
+pub use crate::TypedHeader;
 
 pub(crate) fn has_content_type<B>(
     req: &RequestParts<B>,
     expected_content_type: &mime::Mime,
-) -> Result<bool, HeadersAlreadyExtracted> {
-    let content_type = if let Some(content_type) = req
-        .headers()
-        .ok_or_else(HeadersAlreadyExtracted::default)?
-        .get(header::CONTENT_TYPE)
-    {
+) -> bool {
+    let content_type = if let Some(content_type) = req.headers().get(header::CONTENT_TYPE) {
         content_type
     } else {
-        return Ok(false);
+        return false;
     };
 
     let content_type = if let Ok(content_type) = content_type.to_str() {
         content_type
     } else {
-        return Ok(false);
+        return false;
     };
 
-    Ok(content_type.starts_with(expected_content_type.as_ref()))
+    content_type.starts_with(expected_content_type.as_ref())
 }
 
 pub(crate) fn take_body<B>(req: &mut RequestParts<B>) -> Result<B, BodyAlreadyExtracted> {
